__all__ = [
    "KEYWORD_PATTERN",
    "TRUE_PATTERN",
    "FALSE_PATTERN",
    "NULL_PATTERN",
    "IDENTIFIER_PATTERN",
    "DOCSTRING_PATTERN",
    "MODULE_PATTERN",
    "STRING_PATTERN",
    "NUMBER_PATTERN",
    "RESOURCE_LOCATION_PATTERN",
]


import keyword

KEYWORD_PATTERN: str = "|".join(rf"\b{kw}\b" for kw in keyword.kwlist)

TRUE_PATTERN: str = r"\b[tT]rue\b"
FALSE_PATTERN: str = r"\b[fF]alse\b"
NULL_PATTERN: str = r"\b(?:null|None)\b"
IDENTIFIER_PATTERN: str = rf"(?!_bolt_|{KEYWORD_PATTERN})[a-zA-Z_][a-zA-Z0-9_]*\b"
MODULE_PATTERN: str = rf"{IDENTIFIER_PATTERN}(?:\.{IDENTIFIER_PATTERN})*"
DOCSTRING_PATTERN: str = r'r?"""(?:\\.|[^\\])*?"""' "|" r"r?'''(?:\\.|[^\\])*?'''"
STRING_PATTERN: str = r'r?"(?:\\.|[^\\\n])*?"' "|" r"r?'(?:\\.|[^\\\n])*?'"
NUMBER_PATTERN: str = r"(?:0|[1-9]\d*)(?:\.\d+)?(?:[eE][+-]?\d+)?\b"
RESOURCE_LOCATION_PATTERN: str = r"(?:\.\./|\./|[0-9a-z_\-\.]+:)[0-9a-z_./-]+"



__all__ = [
    "MemoHandler",
    "MemoRegistry",
    "MemoFileIndex",
    "MemoStorage",
    "MemoInvocation",
    "MemoState",
]


import logging
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from typing import (
    Any,
    Callable,
    ClassVar,
    Generic,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from uuid import UUID

from beet import (
    Cache,
    CachePin,
    Container,
    Context,
    DataPack,
    Generator,
    MultiCache,
    ResourcePack,
)
from beet.core.utils import log_time, remove_path
from mecha import AstChildren, AstRoot, DiagnosticCollection, Mecha, rule

from .ast import AstMemo, AstMemoResult
from .emit import CommandEmitter
from .utils import dump_pickle, load_pickle

logger = logging.getLogger("memo")


T = TypeVar("T")


@dataclass
class MemoState(Generic[T]):
    load: Callable[[Path], T]
    dump: Callable[[Path, T], Any]
    default: Callable[[], T]

    cached_field_name: str = "cached_{}"

    def __set_name__(self, owner: Type["MemoInvocation"], name: str):
        self.cached_field_name = self.cached_field_name.format(name)

    def __get__(self, obj: "MemoInvocation", objtype: Type["MemoInvocation"]) -> T:
        value = getattr(obj, self.cached_field_name, None)
        if value is None:
            if (
                obj.cached
                and (directory := obj.locate_directory())
                and (path := directory / self.cached_field_name).exists()
            ):
                logger.debug(
                    'Load "%s" for invocation "%s/%s".',
                    self.cached_field_name,
                    obj.parent.memo_id.hex,
                    hex(obj.state_id),
                )
                value = self.load(path)
            else:
                value = self.default()
            setattr(obj, self.cached_field_name, value)
        return value

    def __set__(self, obj: "MemoInvocation", value: T):
        if obj.cached and (directory := obj.locate_directory()):
            logger.debug(
                'Persist "%s" for invocation "%s/%s".',
                self.cached_field_name,
                obj.parent.memo_id.hex,
                hex(obj.state_id),
            )
            directory.mkdir(parents=True, exist_ok=True)
            self.dump(directory / self.cached_field_name, value)
        setattr(obj, self.cached_field_name, value)


class MemoInvocation:
    parent: "MemoStorage"
    state_id: int
    cached: bool
    epoch: int

    def __init__(self, parent: "MemoStorage", state_id: int):
        self.parent = parent
        self.state_id = state_id
        self.cached = False
        self.epoch = 0

    output: ClassVar[MemoState[str]] = MemoState(
        load=Path.read_text,
        dump=Path.write_text,
        default=str,
    )

    resource_pack: ClassVar[MemoState[ResourcePack]] = MemoState(
        load=lambda p: ResourcePack(path=p),
        dump=lambda p, pack: pack.save(path=p),
        default=ResourcePack,
    )

    data_pack: ClassVar[MemoState[DataPack]] = MemoState(
        load=lambda p: DataPack(path=p),
        dump=lambda p, pack: pack.save(path=p),
        default=DataPack,
    )

    bindings: ClassVar[MemoState[Tuple[Any, ...]]] = MemoState(
        load=load_pickle,
        dump=dump_pickle,
        default=tuple,
    )

    def __getstate__(self) -> Tuple["MemoStorage", int, bool, int]:
        return self.parent, self.state_id, self.cached, self.epoch

    def __setstate__(self, state: Tuple["MemoStorage", int, bool, int]):
        self.parent, self.state_id, self.cached, self.epoch = state

    def locate_directory(self) -> Optional[Path]:
        if path := self.parent.locate_directory():
            return path / hex(self.state_id)
        return None


class MemoStorage(Container[Tuple[Any, ...], MemoInvocation]):
    parent: "MemoFileIndex"
    memo_id: UUID
    memo_lineno: int
    invocation_counter: int

    def __init__(self, parent: "MemoFileIndex", memo_id: UUID, lineno: int = -1):
        super().__init__()
        self.parent = parent
        self.memo_id = memo_id
        self.memo_lineno = lineno
        self.invocation_counter = 0

    def __delitem__(self, key: Tuple[Any, ...]):
        if directory := self[key].locate_directory():
            remove_path(directory)
        super().__delitem__(key)

    def missing(self, key: Tuple[Any, ...]) -> MemoInvocation:
        self.invocation_counter += 1
        return MemoInvocation(parent=self, state_id=self.invocation_counter)

    def locate_directory(self) -> Optional[Path]:
        if path := self.parent.path:
            return path.with_name(self.memo_id.hex)
        return None


class MemoFileIndex(Container[Union[Tuple[AstMemo, int], UUID], MemoStorage]):
    filename: str
    path: Optional[Path]

    def __init__(self, filename: str, path: Optional[Path] = None):
        super().__init__()
        self.filename = filename
        self.path = path

    def __getitem__(self, key: Union[Tuple[AstMemo, int], UUID]) -> MemoStorage:
        if isinstance(key, UUID):
            return super().__getitem__(key)

        memo = key[0]

        storage = self.get(memo.persistent_id)
        if storage is None:
            storage = super().__getitem__(key)
            storage.memo_lineno = memo.location.lineno
            self[memo.persistent_id] = storage

        return storage

    def __delitem__(self, key: Union[Tuple[AstMemo, int], UUID]):
        if isinstance(key, UUID):
            super().__delitem__(key)
            return
        if directory := self[key].locate_directory():
            remove_path(directory)
        del self[key[0].persistent_id]
        super().__delitem__(key)

    def missing(self, key: Union[Tuple[AstMemo, int], UUID]) -> MemoStorage:
        if isinstance(key, UUID):
            return super().missing(key)
        return MemoStorage(self, key[0].persistent_id)

    def clear(self):
        super().clear()
        if self.path:
            remove_path(self.path)

    def garbage_collect(self, gc_cutoff: int) -> Iterator[Tuple[UUID, int]]:
        keep_alive: Set[UUID] = set()

        for storage_key, storage in self.items():
            if isinstance(storage_key, UUID):
                for variables, invocation in list(storage.items()):
                    if not invocation.cached or invocation.epoch < gc_cutoff:
                        yield storage.memo_id, invocation.state_id
                        del storage[variables]
                if storage:
                    keep_alive.add(storage_key)

        for storage_key in list(self):
            if (
                isinstance(storage_key, tuple)
                and storage_key[0].persistent_id not in keep_alive
            ):
                del self[storage_key]

        for storage_key in list(self):
            if isinstance(storage_key, UUID) and storage_key not in keep_alive:
                del self[storage_key]


class MemoRegistry(Container[str, MemoFileIndex]):
    cache: Optional[Cache]

    epoch_counter: ClassVar[CachePin[int]] = CachePin("epoch_counter", 0)
    gc_epoch: ClassVar[CachePin[int]] = CachePin("gc_epoch", 0)
    gc_interval: ClassVar[CachePin[int]] = CachePin("gc_interval", 50)
    gc_max_age: ClassVar[CachePin[int]] = CachePin("gc_max_age", 30)

    def __init__(self, arg: Union[Context, MultiCache[Any], Optional[Cache]] = None):
        super().__init__()
        arg = arg.cache if isinstance(arg, Context) else arg
        self.cache = arg["memo"] if isinstance(arg, MultiCache) else arg

    def __setitem__(self, key: str, value: MemoFileIndex):
        super().__setitem__(key, value)
        if self.cache:
            self.cache.json.setdefault("files", {})[key] = str(value.path)

    def __delitem__(self, key: str):
        self[key].clear()
        if self.cache:
            del self.cache.json.setdefault("files", {})[key]
        super().__delitem__(key)

    def __iter__(self) -> Iterator[str]:
        if self.cache:
            for key in self.cache.json.setdefault("files", {}):
                yield key

    def missing(self, key: str) -> MemoFileIndex:
        if not self.cache:
            return MemoFileIndex(key)

        index_path = self.cache.get_path(f"{key}-memo")
        if index_path.is_file():
            if isinstance(memo_file_index := load_pickle(index_path), MemoFileIndex):
                memo_file_index.path = index_path
                return memo_file_index

        return MemoFileIndex(key, index_path)

    def flush(self):
        for memo_file_index in self.values():
            if memo_file_index.path:
                dump_pickle(memo_file_index.path, memo_file_index)

    @property
    def gc_cutoff(self) -> int:
        return self.epoch_counter - self.gc_max_age

    def garbage_collect(self, *keys: str) -> Iterator[Tuple[UUID, int]]:
        for key in keys:
            file_index = self[key]
            yield from file_index.garbage_collect(self.gc_cutoff)
            if not file_index:
                del self[key]
        self.gc_epoch = self.epoch_counter


@dataclass
class MemoHandler:
    mc: Mecha
    registry: MemoRegistry
    generate: Optional[Generator] = None

    def __post_init__(self):
        self.mc.serialize.extend(serialize_memo_result)

    def restore(self, emit: CommandEmitter, invocation: MemoInvocation):
        if self.generate is None:
            return

        invocation.epoch = self.registry.epoch_counter

        if output := invocation.output:
            emit.commands.append(AstMemoResult(serialized=output))

        self.generate.assets.merge(invocation.resource_pack)
        self.generate.data.merge(invocation.data_pack)

    @contextmanager
    def record(
        self,
        emit: CommandEmitter,
        invocation: MemoInvocation,
        invocation_path: str,
        name: Optional[str] = None,
    ):
        if self.generate is None or name is None:
            yield
            return

        database = self.mc.database
        source = database.index[name]
        compilation_unit = database[source]

        with self.generate.draft() as draft, draft.push():
            with emit.scope() as commands:
                yield

            root = AstRoot(commands=AstChildren(commands))

            diagnostics = DiagnosticCollection()

            previous_session = database.session
            previous_queue = database.queue
            previous_step = database.step
            previous_current = database.current
            try:
                database.session = set()
                database.queue = []
                function = self.mc.compile(
                    root,
                    filename=compilation_unit.filename,
                    resource_location=invocation_path,
                    within=draft.data,
                    report=diagnostics,
                    initial_step=database.step + 1,
                )
            finally:
                previous_session.update(database.session)
                database.session = previous_session
                database.queue = previous_queue
                database.step = previous_step
                database.current = previous_current

            if output := function.text.rstrip():
                emit.commands.append(AstMemoResult(serialized=output))

            database[function].source = compilation_unit.source

            if not diagnostics.exceptions:
                invocation.epoch = self.registry.epoch_counter
                invocation.cached = True

                if output:
                    invocation.output = output
                if draft.assets:
                    invocation.resource_pack = draft.assets
                if draft.data:
                    invocation.data_pack = draft.data

    def finalize(self):
        self.registry.epoch_counter += 1
        next_gc = self.registry.gc_epoch + self.registry.gc_interval
        if self.registry.epoch_counter >= next_gc:
            with log_time("GC bolt memo."):
                for memo_id, state_id in self.registry.garbage_collect(*self.registry):
                    logger.debug('GC "%s/%s".', memo_id.hex, hex(state_id))
        self.registry.flush()


@rule(AstMemoResult)
def serialize_memo_result(node: AstMemoResult, result: List[str]):
    result.append(node.serialized)



__all__ = [
    "LoopInfo",
    "loop_info",
]


from dataclasses import dataclass, field
from typing import (
    Any,
    Generic,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from beet.core.utils import SENTINEL_OBJ, Sentinel

T = TypeVar("T")
U = TypeVar("U")


def loop_info(iterable: Iterable[T]) -> "LoopInfo[T]":
    return LoopInfo(iterable, iter(iterable))


@dataclass
class LoopInfo(Generic[T]):
    iterable: Iterable[T]
    iterator: Iterator[T]

    index: int = -1
    accumulator: List[T] = field(default_factory=list)

    def __iter__(self) -> "LoopInfo[T]":
        return self

    def __next__(self) -> Tuple["LoopInfo[T]", T]:
        while self.index + 1 >= len(self.accumulator):
            self.accumulator.append(next(self.iterator))
        self.index += 1
        return self, self.current

    @overload
    def peek(self, offset: int) -> Optional[T]: ...

    @overload
    def peek(self, offset: int, default: U) -> Union[T, U]: ...

    def peek(self, offset: int, default: Any = None) -> Any:
        index = self.index + offset

        if index < 0:
            return default

        try:
            while index >= len(self.accumulator):
                self.accumulator.append(next(self.iterator))
        except StopIteration:
            return default

        return self.accumulator[index]

    @property
    def before(self) -> Optional[T]:
        return self.peek(-1)

    @property
    def current(self) -> T:
        return self.accumulator[self.index]

    @property
    def after(self) -> Optional[T]:
        return self.peek(1)

    @property
    def count(self) -> int:
        try:
            return len(self.iterable)  # type: ignore
        except TypeError:
            self.accumulator.extend(self.iterator)
            return len(self.accumulator)

    @property
    def first(self) -> bool:
        return self.index == 0

    @property
    def last(self) -> bool:
        return isinstance(self.peek(1, SENTINEL_OBJ), Sentinel)

    def cycle(self, arg: U, *args: U) -> U:
        args = (arg,) + args
        return args[self.index % len(args)]



"""Plugin that sandboxes the bolt runtime."""

__all__ = [
    "Sandbox",
    "SandboxedAttributeHandler",
    "SandboxedImportModule",
    "SecurityError",
    "public_attrs",
]


import bisect
import collections
import copy
import heapq
import itertools
import json
import math
import random
from collections import defaultdict, deque
from dataclasses import dataclass, fields, is_dataclass
from typing import Any, Callable, DefaultDict, Set, Type, Union, get_type_hints

from beet import Context, TreeData, TreeNode
from mecha import MechaError

from bolt import LoopInfo, Runtime
from bolt.utils import internal


class SecurityError(MechaError):
    """Raised when the sandboxed runtime attempts to perform a forbidden operation."""


def beet_default(ctx: Context):
    ctx.require("beet.contrib.template_sandbox")
    sandbox = ctx.inject(Sandbox)
    sandbox.allow_basics()
    sandbox.allow_pipeline_context()
    sandbox.activate()


def public_attrs(obj: Any) -> Set[str]:
    """List public object attributes."""
    attrs = dir(obj)
    obj_type = obj if isinstance(obj, type) else type(obj)
    if is_dataclass(obj_type):
        attrs.extend(f.name for f in fields(obj_type))
    else:
        attrs.extend(k for t in type.mro(obj_type) for k in get_type_hints(t))
    return {name for name in attrs if not name.startswith("_")}


class Sandbox:
    """Bolt sandbox."""

    runtime: Runtime
    active: bool

    allowed_builtins: Set[str]
    allowed_imports: Set[str]
    allowed_obj_attrs: DefaultDict[Any, Set[str]]
    allowed_type_attrs: DefaultDict[Type[Any], Set[str]]

    def __init__(self, ctx: Union[Context, Runtime]):
        if isinstance(ctx, Context):
            self.runtime = ctx.inject(Runtime)
        else:
            self.runtime = ctx

        self.active = False

        self.allowed_builtins = set()
        self.allowed_imports = set()
        self.allowed_obj_attrs = defaultdict(set)
        self.allowed_type_attrs = defaultdict(set)

    def allow_basics(self):
        self.allowed_builtins |= {
            "abs",
            "all",
            "any",
            "ascii",
            "bin",
            "bool",
            "callable",
            "chr",
            "dict",
            "divmod",
            "enumerate",
            "filter",
            "float",
            "frozenset",
            "hasattr",
            "hash",
            "hex",
            "int",
            "isinstance",
            "issubclass",
            "iter",
            "len",
            "list",
            "map",
            "max",
            "min",
            "next",
            "object",
            "oct",
            "ord",
            "pow",
            "print",
            "range",
            "repr",
            "reversed",
            "round",
            "set",
            "slice",
            "sorted",
            "str",
            "sum",
            "tuple",
            "type",
            "zip",
        }

        self.allowed_imports |= {
            "json",
            "math",
            "collections",
            "itertools",
            "bisect",
            "heapq",
            "copy",
            "random",
        }

        self.allowed_obj_attrs |= {
            dict: {"fromkeys"},
            defaultdict: {"fromkeys"},
            json: {"loads", "dumps"},
            math: public_attrs(math),
            collections: {"deque", "defaultdict"},
            itertools: public_attrs(itertools),
            bisect: public_attrs(bisect),
            heapq: public_attrs(heapq),
            copy: {"copy", "deepcopy"},
            random: public_attrs(random),
        }

        self.allowed_type_attrs |= {
            str: public_attrs(str)
            ^ {
                "maketrans",
                "translate",
                "encode",
                "format",
                "format_map",
            },
            tuple: public_attrs(tuple),
            list: public_attrs(list),
            dict: public_attrs(dict),
            set: public_attrs(set),
            TreeNode: public_attrs(TreeNode),
            TreeData: public_attrs(TreeData),
            LoopInfo: public_attrs(LoopInfo),
            deque: public_attrs(deque),
            defaultdict: public_attrs(defaultdict),
        }

    def allow_pipeline_context(self):
        self.allowed_type_attrs[Context] |= {
            "project_id",
            "project_name",
            "project_description",
            "project_author",
            "project_version",
            "project_root",
            "minecraft_version",
        }

    def activate(self):
        """Activate the sandbox."""
        if self.active:
            return
        self.active = True

        self.runtime.builtins &= self.allowed_builtins

        get_attribute_handler = self.runtime.helpers["get_attribute_handler"]

        self.runtime.helpers.update(
            get_attribute_handler=lambda obj: SandboxedAttributeHandler(  # type: ignore
                obj=obj,
                handler=get_attribute_handler(obj),
                sandbox=self,
            ),
            import_module=SandboxedImportModule(
                sandbox=self,
                import_module=self.runtime.helpers["import_module"],
            ),
        )


@dataclass
class SandboxedAttributeHandler:
    """Sandboxed attribute handler helper."""

    obj: Any
    handler: Any
    sandbox: Sandbox

    @internal
    def __getitem__(self, attr: str) -> Any:
        if not hasattr(self.obj, attr):
            return self.handler[attr]

        try:
            if allowed := self.sandbox.allowed_obj_attrs.get(self.obj):
                if attr in allowed:
                    return self.handler[attr]
        except TypeError:
            pass

        for cls in type.mro(type(self.obj)):
            if allowed := self.sandbox.allowed_type_attrs.get(cls):
                if attr in allowed:
                    return self.handler[attr]

        raise SecurityError(f"Access forbidden attribute {attr!r} of {type(self.obj)}.")


@dataclass
class SandboxedImportModule:
    """Sandboxed import_module helper."""

    sandbox: Sandbox
    import_module: Callable[[str], Any]

    @internal
    def __call__(self, name: str) -> Any:
        if name in self.sandbox.allowed_imports:
            return self.import_module(name)
        raise SecurityError(f"Access forbidden module {name!r}.")



__all__ = [
    "bolt",
    "beet_default",
]


from beet import Context, ListOption, configurable
from pydantic.v1 import BaseModel

from .runtime import Runtime


class BoltOptions(BaseModel):
    entrypoint: ListOption[str] = ListOption()
    prelude: ListOption[str] = ListOption()


def beet_default(ctx: Context):
    ctx.require(bolt)


@configurable(validator=BoltOptions)
def bolt(ctx: Context, opts: BoltOptions):
    """Plugin for configuring bolt."""
    runtime = ctx.inject(Runtime)
    runtime.evaluate.add_entrypoint(opts.entrypoint.entries())
    runtime.modules.add_prelude(opts.prelude.entries())



"""Run deferred callbacks."""

__all__ = [
    "Defer",
    "DeferHandler",
    "AstDefer",
    "AstDeferCommand",
    "defer",
]


from dataclasses import dataclass
from typing import Any, Callable, Generator, List, TypeVar

from beet import Context
from beet.core.utils import required_field
from mecha import (
    AstChildren,
    AstCommandSentinel,
    AstNode,
    AstRoot,
    Dispatcher,
    Mecha,
    MutatingReducer,
    rule,
)

from bolt import Runtime

CallbackType = TypeVar("CallbackType", bound=Callable[[], Any])


def beet_default(ctx: Context):
    ctx.require(defer)


def defer(ctx: Context):
    """Plugin that makes the defer utility available globally."""
    runtime = ctx.inject(Runtime)
    defer = ctx.inject(Defer)
    runtime.expose("defer", defer)


@dataclass(frozen=True, slots=True)
class AstDefer(AstNode):
    """Ast defer node."""

    callback: Callable[[], Any] = required_field()


@dataclass(frozen=True, slots=True)
class AstDeferCommand(AstCommandSentinel):
    """Ast defer command node.

    Same as AstDefer but allows the node to pass as a regular AstCommand.
    """

    callback: Callable[[], Any] = required_field()


@dataclass
class DeferHandler(MutatingReducer):
    runtime: Runtime = required_field()

    @rule(AstDefer)
    @rule(AstDeferCommand)
    def defer(self, node: AstDefer | AstDeferCommand) -> Generator[AstNode, Any, Any]:
        with self.runtime.modules.error_handler(
            "Deferred callback raised an exception."
        ):
            result = node.callback()

        if isinstance(result, AstNode):
            result = yield result
        elif isinstance(result, AstChildren):
            children: List[Any] = []
            for child in result:  # type: ignore
                deferred_child = yield child
                children.append(deferred_child)
            result = AstChildren(children)

        return result


class Defer:
    runtime: Runtime
    handler: Dispatcher[AstRoot]

    def __init__(self, ctx: Context):
        self.runtime = ctx.inject(Runtime)
        self.handler = DeferHandler(runtime=self.runtime)
        mc = ctx.inject(Mecha)
        mc.steps.insert(mc.steps.index(self.runtime.evaluate) + 1, self.handler)

    def __call__(self, callback: CallbackType) -> CallbackType:
        self.runtime.commands.append(
            AstDeferCommand(
                callback=lambda: AstChildren(self.runtime.capture_output(callback))
            )
        )
        return callback



__version__ = "0.49.1"


from .ast import *
from .codegen import *
from .emit import *
from .helpers import *
from .loop_info import *
from .memo import *
from .module import *
from .parse import *
from .plugin import *
from .runtime import *
from .semantics import *



__all__ = [
    "Module",
    "AssetModule",
    "CompiledModule",
    "ModuleManager",
    "ModuleCacheBackend",
    "MacroLibrary",
    "CodegenResult",
    "UnusableCompilationUnit",
    "SilentCompilationInterrupt",
]


import logging
import marshal
from contextlib import contextmanager
from dataclasses import dataclass, field
from io import BufferedReader, BufferedWriter
from pathlib import Path
from types import CodeType
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Protocol,
    Set,
    Tuple,
    Union,
    cast,
)

from beet import (
    BubbleException,
    Cache,
    DataPack,
    NamespaceFileScope,
    ResourcePack,
    TextFile,
    TextFileBase,
)
from beet.core.utils import JsonDict, extra_field, import_from_string, required_field
from mecha import (
    AstCacheBackend,
    AstChildren,
    AstNode,
    AstResourceLocation,
    AstRoot,
    CompilationDatabase,
    CompilationError,
    CompilationUnit,
    DiagnosticCollection,
    DiagnosticError,
    Dispatcher,
    MechaError,
)
from tokenstream import InvalidSyntax

from .ast import AstImportedItem, AstImportedMacro, AstMacro, AstPrelude
from .semantics import LexicalScope
from .utils import internal, rewrite_traceback

logger = logging.getLogger("bolt")


class UnusableCompilationUnit(MechaError):
    """Raised when a compilation unit can not be used to instantiate a module."""

    message: str
    compilation_unit: CompilationUnit

    def __init__(self, message: str, compilation_unit: CompilationUnit) -> None:
        super().__init__(message, compilation_unit)
        self.message = message
        self.compilation_unit = compilation_unit

    def __str__(self) -> str:
        return self.message


class SilentCompilationInterrupt(DiagnosticCollection):
    """Raised for interrupting the compilation when an unusable compilation unit already reported its own diagnostics."""


class Module(TextFile):
    """Class representing a bolt module."""

    scope: ClassVar[NamespaceFileScope] = {
        0: ("modules",),
        48: ("module",),
    }
    extension: ClassVar[str] = ".bolt"


class AssetModule(TextFile):
    """Class representing a bolt module in a resource pack."""

    scope: ClassVar[NamespaceFileScope] = {
        0: ("modules",),
        34: ("module",),
    }
    extension: ClassVar[str] = ".bolt"


MacroLibrary = Dict[str, Dict[Tuple[str, AstMacro], Optional[Tuple[str, str]]]]


@dataclass
class CompiledModule:
    """Class holding the state of a compiled module."""

    ast: AstRoot
    code: Optional[CodeType]
    python: Optional[str]
    refs: List[Any]
    dependencies: Set[str]
    prelude_imports: List[AstPrelude]
    macros: MacroLibrary
    lexical_scope: LexicalScope
    output: Optional[str]
    resource_location: Optional[str]
    globals: Set[str]
    namespace: JsonDict = field(default_factory=dict)
    executing: bool = False
    executed: bool = False
    execution_hooks: List[Callable[[], Any]] = field(default_factory=list)
    execution_index: int = 0


@dataclass(frozen=True)
class CodegenResult:
    source: Optional[str] = None
    output: Optional[str] = None
    refs: List[Any] = field(default_factory=list)
    dependencies: Set[str] = field(default_factory=set)
    prelude_imports: List[AstPrelude] = field(default_factory=list)
    macros: MacroLibrary = field(default_factory=dict)


class ModuleParseCallback(Protocol):
    """Callback required by the module manager for parsing."""

    def __call__(
        self,
        source: TextFileBase[Any],
        *,
        filename: Optional[str],
        resource_location: Optional[str],
    ) -> AstRoot: ...


@dataclass
class ModuleManager(Mapping[TextFileBase[Any], CompiledModule]):
    """Container for managing bolt modules."""

    directory: Path = extra_field()
    database: CompilationDatabase = extra_field()
    codegen: Dispatcher[CodegenResult] = extra_field()
    parse_callback: ModuleParseCallback = extra_field()
    cache: Optional[Cache] = extra_field(default=None)

    registry: Dict[TextFileBase[Any], CompiledModule] = extra_field(
        default_factory=dict
    )
    stack: List[CompiledModule] = extra_field(default_factory=list)
    parse_stack: List[TextFileBase[Any]] = extra_field(default_factory=list)
    parse_scopes: Dict[TextFileBase[Any], LexicalScope] = extra_field(
        default_factory=dict
    )
    globals: JsonDict = extra_field(default_factory=dict)
    builtins: Set[str] = extra_field(default_factory=set)
    prelude: Dict[str, Dict[Optional[Union[ResourcePack, DataPack]], AstPrelude]] = (
        extra_field(default_factory=dict)
    )

    execution_count: int = 0

    @property
    def current(self) -> CompiledModule:
        """Return the currently executing module."""
        if not self.stack:
            raise ValueError("No module currently executing.")
        return self.stack[-1]

    @property
    def current_path(self) -> str:
        """Return the path corresponding to the currently executing module."""
        if path := self.current.resource_location:
            return path
        raise ValueError(
            "Currently executing module has no associated resource location."
        )

    def lookup(self, path: str) -> TextFileBase[Any]:
        """Lookup the associated source file in the compilation database.

        If we're currently in an overlay, fall back to the source file in the main
        data pack.
        """
        if current := self.database.index.get(path):
            return current

        current_pack = self.database[self.database.current].pack
        if current_pack is None or current_pack.overlay_parent is None:
            raise KeyError(path)

        return self.database.indices[current_pack.overlay_parent][path]

    def match_ast(
        self,
        node: AstRoot,
        current: Optional[Union[TextFileBase[Any], str]] = None,
    ) -> Tuple[CompilationUnit, CompiledModule]:
        """Return the compilation unit and executable module for the current ast."""
        if not current:
            current = self.database.current
        elif isinstance(current, str):
            current = self.lookup(current)

        compilation_unit = self.database[current]
        compilation_unit.ast = node

        return compilation_unit, self[current]

    def __getitem__(self, current: Union[TextFileBase[Any], str]) -> CompiledModule:
        if isinstance(current, str):
            current = self.lookup(current)

        compilation_unit = self.database[current]
        name = compilation_unit.resource_location or "<unknown>"

        if module := self.registry.get(current):
            if (
                module.executed
                or not compilation_unit.ast
                or module.ast is compilation_unit.ast
            ):
                return module
            logger.warning('Code generation due to ast update "%s".', name)

        elif not compilation_unit.ast:
            previous = self.database.current
            self.database.current = current
            try:
                compilation_unit.ast = self.parse_callback(
                    current,
                    filename=compilation_unit.filename,
                    resource_location=compilation_unit.resource_location,
                )
            except DiagnosticError as exc:
                raise UnusableCompilationUnit(
                    "Parsing failed.", compilation_unit
                ) from exc
            finally:
                self.database.current = previous
            return self[current]

        else:
            logger.debug('Code generation "%s".', name)

        result = self.codegen(compilation_unit.ast)

        if result.source and result.output:
            code = compile(
                source=result.source,
                filename=(
                    str(self.directory / compilation_unit.filename)
                    if compilation_unit.filename
                    else name
                ),
                mode="exec",
            )

        else:
            code = None

        module = CompiledModule(
            ast=compilation_unit.ast,
            code=code,
            python=result.source,
            refs=result.refs,
            dependencies=result.dependencies,
            prelude_imports=result.prelude_imports,
            macros=result.macros,
            lexical_scope=self.parse_scopes.get(current) or LexicalScope(),
            output=result.output,
            resource_location=compilation_unit.resource_location,
            globals=set(self.globals),
        )
        self.registry[current] = module

        return module

    def __iter__(self) -> Iterator[TextFileBase[Any]]:
        return iter(self.database)

    def __len__(self) -> int:
        return len(self.database)

    def get(self, current: Union[TextFileBase[Any], str]) -> Optional[CompiledModule]:
        """Return executable module if exists."""
        if isinstance(current, str):
            try:
                current = self.lookup(current)
            except KeyError:
                return None

        if current not in self.database:
            return None

        return self[current]

    @internal
    def eval(self, module: CompiledModule) -> AstRoot:
        """Run the module and return the output ast."""
        if not module.executed:
            module.executed = True
            for hook in module.execution_hooks:
                hook()

        if not module.code or not module.output:
            return module.ast
        if module.output in module.namespace:
            return module.namespace[module.output]

        if module.executing:
            raise ValueError("Import cycle detected.")

        logger.debug('Evaluate module "%s".', module.resource_location or "<unknown>")

        module.namespace.update(self.globals)
        module.namespace["_bolt_refs"] = module.refs
        module.namespace["__name__"] = module.resource_location
        module.namespace["__file__"] = module.code.co_filename

        self.stack.append(module)
        module.executing = True

        try:
            exec(module.code, module.namespace)
        finally:
            module.executing = False
            self.stack.pop()
            self.execution_count += 1
            module.execution_index = self.execution_count

        return module.namespace[module.output]

    @contextmanager
    @internal
    def error_handler(self, message: str, resource_location: Optional[str] = None):
        """Error handler for running module code."""
        try:
            yield
        except (BubbleException, InvalidSyntax, SilentCompilationInterrupt):
            raise
        except UnusableCompilationUnit as exc:
            if not exc.compilation_unit.diagnostics.error:
                message = "Failed to instantiate module."
                if resource_location:
                    message += f" ({resource_location})"
                raise CompilationError(message) from rewrite_traceback(exc)
            raise SilentCompilationInterrupt() from None
        except Exception as exc:
            if resource_location:
                message += f" ({resource_location})"
            raise CompilationError(message) from rewrite_traceback(exc)

    @contextmanager
    def parse_push(self, current: TextFileBase[Any]):
        """Push the current file onto the parse stack."""
        if current in self.parse_stack:
            stack = " -> ".join(
                f'"{self.database[entry].resource_location or "<unknown>"}"'
                for entry in self.parse_stack + [current]
            )
            raise UnusableCompilationUnit(
                f"Cyclic parsing dependency {stack}.",
                compilation_unit=self.database[current],
            )

        self.parse_stack.append(current)

        try:
            yield
        finally:
            self.parse_stack.pop()

    def add_prelude(self, *args: Union[str, Iterable[str]]):
        """Add modules to the prelude."""
        for arg in args:
            if isinstance(arg, str):
                arg = [arg]
            for resource_location in arg:
                self.prelude.setdefault(resource_location, {})

    def export_prelude(self, module: CompiledModule) -> Optional[AstPrelude]:
        """Export local variables and macros from a given module into a prelude import."""
        if not module.resource_location:
            return None

        arguments: List[AstNode] = [
            AstResourceLocation.from_value(module.resource_location)
        ]

        for name, variable in module.lexical_scope.variables.items():
            if variable.storage == "local":
                arguments.append(AstImportedItem(name=name))

        for overloads in module.macros.values():
            for name, macro in overloads:
                arguments.append(AstImportedMacro(name=name, declaration=macro))

        return AstPrelude(arguments=AstChildren(arguments))

    def load_prelude(self) -> List[AstPrelude]:
        """Load the prelude imports."""
        current_pack = self.database[self.database.current].pack
        current_prelude = self.prelude

        prelude_imports: List[AstPrelude] = []

        for resource_location, pack_prelude in list(current_prelude.items()):
            prelude = pack_prelude.get(current_pack)
            if not prelude:
                self.prelude = {}
                try:
                    module = self[resource_location]
                except UnusableCompilationUnit:
                    pass
                else:
                    prelude = self.export_prelude(module)
                finally:
                    self.prelude = current_prelude

            if not prelude:
                prelude = AstPrelude.placeholder(resource_location)

            pack_prelude[current_pack] = prelude
            prelude_imports.append(prelude)

        return prelude_imports


@dataclass
class ModuleCacheBackend(AstCacheBackend):
    """Cache backend that also restores the generated modules."""

    bolt_version: str = import_from_string("bolt.__version__")

    modules: ModuleManager = required_field(repr=False, hash=False, compare=False)

    def load_data(self, f: BufferedReader) -> JsonDict:
        data = super().load_data(f)

        if data["bolt"] != self.bolt_version:
            raise ValueError("Version mismatch.")
        if data["globals"] != set(self.modules.globals):
            raise ValueError("Globals mismatch.")
        if data["prelude_imports"] != self.modules.load_prelude():
            raise ValueError("Prelude mismatch.")

        for group, exports in cast(MacroLibrary, data["macros"]).items():
            for (_, macro), extern in exports.items():
                if extern:
                    resource_location, name = extern
                    macros = self.modules[resource_location].macros
                    if group not in macros or (name, macro) not in macros[group]:
                        raise ValueError("Macro mismatch.")

        return data

    def dump_data(self, data: JsonDict, f: BufferedWriter):
        data["bolt"] = self.bolt_version
        super().dump_data(data, f)

    def load(self, f: BufferedReader) -> AstRoot:
        data = self.load_data(f)

        self.modules.registry[self.modules.database.current] = CompiledModule(
            ast=data["ast"],
            code=marshal.load(f),
            python=data["codegen"],
            refs=data["refs"],
            dependencies=data["dependencies"],
            prelude_imports=data["prelude_imports"],
            macros=data["macros"],
            lexical_scope=data["lexical_scope"],
            output=data["output"],
            resource_location=data["resource_location"],
            globals=data["globals"],
        )

        for dependency in data["dependencies"]:
            self.modules.get(dependency)

        return data["ast"]

    def dump(self, node: AstRoot, f: BufferedWriter):
        _, module = self.modules.match_ast(node)

        self.dump_data(
            {
                "ast": module.ast,
                "codegen": module.python,
                "refs": module.refs,
                "dependencies": module.dependencies,
                "prelude_imports": module.prelude_imports,
                "macros": module.macros,
                "lexical_scope": module.lexical_scope,
                "output": module.output,
                "resource_location": module.resource_location,
                "globals": module.globals,
            },
            f,
        )

        marshal.dump(module.code, f)



import os
from pathlib import Path
from pprint import pformat
from typing import Iterable, Optional, Tuple

import click
from beet import ErrorMessage, Project
from beet.core.utils import format_directory
from beet.toolchain.cli import beet, message_fence
from mecha import AstCacheBackend

from bolt import AstMemo, MemoRegistry, MemoStorage

pass_project = click.make_pass_decorator(Project)  # type: ignore


@beet.command()
@pass_project
@click.argument("filename", type=click.Path(exists=True, dir_okay=False))
def codegen(project: Project, filename: str):
    """Inspect cached bolt codegen."""
    source_path = Path(filename).resolve()
    ast_path = project.cache["mecha"].get_path(f"{source_path}-ast")

    try:
        with ast_path.open("rb") as f:
            click.echo(AstCacheBackend().load_data(f)["codegen"], nl=False)
    except Exception as exc:
        raise ErrorMessage(f'No cached codegen for "{filename}".') from exc


@beet.command()
@pass_project
@click.option(
    "-c",
    "--clear",
    is_flag=True,
    help="Clear cached memo.",
)
@click.option(
    "--gc",
    is_flag=True,
    help="Garbage-collect unused invocations.",
)
@click.option(
    "--gc-interval",
    type=int,
    help="Set garbage collection interval.",
)
@click.option(
    "--gc-max-age",
    type=int,
    help="Set garbage collection max age.",
)
@click.argument(
    "filenames",
    nargs=-1,
)
def memo(
    project: Project,
    clear: bool,
    gc: bool,
    gc_interval: Optional[int],
    gc_max_age: Optional[int],
    filenames: Tuple[str, ...],
):
    """Inspect and manage bolt memo storage."""
    memo_registry = MemoRegistry(project.cache)

    if gc_interval is not None or gc_max_age is not None:
        if clear or gc or filenames:
            raise click.UsageError(
                "GC settings must be provided without any other argument.",
            )
        with message_fence("Updating GC settings..."):
            if gc_interval is not None:
                memo_registry.gc_interval = gc_interval
                click.echo(
                    f"Set garbage collection interval to {gc_interval} epoch{(gc_interval > 1) * 's'}.\n"
                )
            if gc_max_age is not None:
                memo_registry.gc_max_age = gc_max_age
                click.echo(
                    f"Set garbage collection max age to {gc_max_age} epoch{(gc_max_age > 1) * 's'}.\n"
                )
            project.cache.flush()
            return

    keys = [
        str(path) if (path := Path(filename).resolve()).is_file() else filename
        for filename in filenames
    ] or list(memo_registry.keys())

    with message_fence(
        "Clearing memo..."
        if clear
        else "Running garbage collection..." if gc else "Inspecting memo..."
    ):
        if not keys:
            click.echo("The memo registry is empty.\n")
            return

        if clear:
            for key in keys:
                click.echo(
                    f'Clear memo for "{os.path.relpath(key, project.directory)}".\n'
                )
                del memo_registry[key]
            memo_registry.flush()
            project.cache.flush()
            return

        if gc:
            for memo_id, state_id in memo_registry.garbage_collect(*keys):
                click.echo(f'GC "{memo_id.hex}/{hex(state_id)}".\n')
            memo_registry.flush()
            project.cache.flush()
            return

        n = len(keys)
        click.echo(f"Looking up {n} file{(n > 1) * 's'} in the registry.\n")

        for key in keys:
            memo_file_index = memo_registry[key]
            if not memo_file_index:
                click.echo(
                    f'No entries associated with "{os.path.relpath(key, project.directory)}".\n'
                )
                continue

            click.echo(
                os.path.relpath(key, project.directory) if os.path.isabs(key) else key
            )

            for key, storage in memo_file_index.items():
                if isinstance(key, tuple):
                    click.echo("  |")
                    for line in generate_summary(project.directory, key[0], storage):
                        click.echo(f"  |  {line}".rstrip())

            click.echo()


def generate_summary(
    project_directory: Path,
    memo: AstMemo,
    storage: MemoStorage,
) -> Iterable[str]:
    yield f"line {storage.memo_lineno}:"
    yield f"  id = {storage.memo_id.hex}"

    yield ""
    yield from memo.dump(
        prefix="  ",
        shallow=True,
        exclude={"location", "end_location", "persistent_id"},
    ).splitlines()

    yield ""
    yield "  invocations ="

    if not storage:
        yield "    None"
        return

    for variable, invocation in storage.items():
        for line in pformat(variable).splitlines():
            yield f"    {line}"
        yield f"      cached = {invocation.cached}"
        yield f"      epoch = {invocation.epoch}"

        directory = invocation.locate_directory()

        if directory and directory.is_dir():
            yield f"      directory = {os.path.relpath(directory, project_directory)}"
            for line in format_directory(directory):  # type: ignore
                yield f"        {line}"
        else:
            yield f"      directory = None"



"""Plugin for emitting generated code."""

__all__ = [
    "DebugCodegenEmitter",
]


from dataclasses import dataclass

from beet import Context
from beet.core.utils import required_field
from mecha import AstRoot, Mecha, Visitor, rule

from bolt import Runtime
from bolt.module import ModuleManager


def beet_default(ctx: Context):
    mc = ctx.inject(Mecha)
    runtime = ctx.inject(Runtime)
    mc.steps[:] = [DebugCodegenEmitter(modules=runtime.modules)]


@dataclass
class DebugCodegenEmitter(Visitor):
    """Visitor that interrupts the compilation process and dumps the generated code."""

    modules: ModuleManager = required_field()

    @rule(AstRoot)
    def debug_codegen(self, node: AstRoot):
        self.modules.database.current.text = self.modules.codegen(node).source or ""
        return None



__all__ = [
    "get_bolt_helpers",
]


from contextlib import ExitStack, contextmanager
from dataclasses import dataclass, replace
from functools import partial, wraps
from importlib import import_module
from typing import Any, Callable, Dict
from uuid import UUID

from mecha import (
    AstAdvancementPredicate,
    AstBool,
    AstChildren,
    AstColor,
    AstCommand,
    AstCoordinate,
    AstGamemode,
    AstGreedy,
    AstItemSlot,
    AstItemSlots,
    AstJson,
    AstJsonObject,
    AstMessage,
    AstNbt,
    AstNbtCompound,
    AstNbtPath,
    AstNode,
    AstNumber,
    AstObjective,
    AstObjectiveCriteria,
    AstPlayerName,
    AstRange,
    AstResourceLocation,
    AstScoreboardSlot,
    AstSortOrder,
    AstString,
    AstSwizzle,
    AstTeam,
    AstTime,
    AstUUID,
    AstVector2,
    AstVector3,
    AstWord,
)
from mecha.contrib.relative_location import resolve_relative_location
from tokenstream import set_location

from .utils import internal


def get_bolt_helpers() -> Dict[str, Any]:
    """Return a collection of helpers used by the generated code."""
    return {
        "replace": replace,
        "missing": object(),
        "exit_stack": ExitStack,
        "children": AstChildren,
        "operator_not": operator_not,
        "operator_in": operator_in,
        "operator_not_in": operator_not_in,
        "branch": branch_driver,
        "loop": loop_driver,
        "get_dup": get_dup,
        "get_rebind": get_rebind,
        "get_attribute_handler": AttributeHandler,
        "import_module": python_import_module,
        "macro_call": macro_call,
        "resolve_formatted_location": resolve_formatted_location,
        "interpolate_bool": converter(AstBool.from_value),
        "interpolate_numeric": converter(AstNumber.from_value),
        "interpolate_coordinate": converter(AstCoordinate.from_value),
        "interpolate_time": converter(AstTime.from_value),
        "interpolate_word": converter(AstWord.from_value),
        "interpolate_phrase": converter(AstString.from_value),
        "interpolate_greedy": converter(AstGreedy.from_value),
        "interpolate_json": converter(AstJson.from_value),
        "interpolate_json_object": JsonObjectConverter(converter(AstJson.from_value)),
        "interpolate_nbt": converter(AstNbt.from_value),
        "interpolate_nbt_compound": NbtCompoundConverter(converter(AstNbt.from_value)),
        "interpolate_nbt_path": converter(AstNbtPath.from_value),
        "interpolate_range": converter(AstRange.from_value),
        "interpolate_resource_location": converter(AstResourceLocation.from_value),
        "interpolate_item_slot": converter(AstItemSlot.from_value),
        "interpolate_item_slots": converter(AstItemSlots.from_value),
        "interpolate_uuid": converter(AstUUID.from_value),
        "interpolate_objective": converter(AstObjective.from_value),
        "interpolate_objective_criteria": converter(AstObjectiveCriteria.from_value),
        "interpolate_scoreboard_slot": converter(AstScoreboardSlot.from_value),
        "interpolate_swizzle": converter(AstSwizzle.from_value),
        "interpolate_team": converter(AstTeam.from_value),
        "interpolate_advancement_predicate": converter(
            AstAdvancementPredicate.from_value
        ),
        "interpolate_color": converter(AstColor.from_value),
        "interpolate_sort_order": converter(AstSortOrder.from_value),
        "interpolate_gamemode": converter(AstGamemode.from_value),
        "interpolate_message": converter(AstMessage.from_value),
        "interpolate_vec2": converter(AstVector2.from_value),
        "interpolate_vec3": converter(AstVector3.from_value),
        "interpolate_entity": EntityConverter(
            uuid_converter=converter(AstUUID.from_value),
            player_name_converter=converter(AstPlayerName.from_value),
        ),
    }


@internal
def operator_not(obj: Any):
    if func := getattr(type(obj), "__not__", None):
        return func(obj)
    return not obj


@internal
def operator_in(item: Any, container: Any):
    if func := getattr(type(item), "__within__", None):
        return func(item, container)
    if func := getattr(type(container), "__contains__", None):
        return func(container, item)
    return item in container


@internal
def operator_not_in(item: Any, container: Any):
    return operator_not(operator_in(item, container))


@contextmanager
@internal
def branch_driver(obj: Any):
    if func := getattr(type(obj), "__branch__", None):
        with func(obj) as condition:
            yield condition
    else:
        yield obj


@contextmanager
@internal
def loop_driver(obj: Any):
    if func := getattr(type(obj), "__loop__", None):
        with func(obj) as cont:
            yield True, cont
    else:
        yield False, True


@internal
def get_dup(obj: Any):
    if func := getattr(type(obj), "__dup__", None):
        return partial(func, obj)
    return None


@internal
def get_rebind(obj: Any):
    if func := getattr(type(obj), "__rebind__", None):
        return partial(func, obj)
    return None


@dataclass
class AttributeHandler:
    obj: Any
    item: bool = False

    @internal
    def __getitem__(self, attr: str) -> Any:
        try:
            return getattr(self.obj, attr)
        except AttributeError as exc:
            try:
                result = self.obj[attr]
                self.item = True
                return result
            except (TypeError, LookupError):
                raise exc from None

    @internal
    def __setitem__(self, attr: str, value: Any):
        try:
            current = self.__getitem__(attr)
        except AttributeError:
            pass
        else:
            if func := getattr(type(current), "__rebind__", None):
                value = func(current, value)
        if self.item:
            self.obj[attr] = value
        else:
            setattr(self.obj, attr, value)

    @internal
    def __delitem__(self, attr: str):
        try:
            self.__getitem__(attr)
        except AttributeError:
            pass
        if self.item:
            del self.obj[attr]
        else:
            delattr(self.obj, attr)


@internal
def python_import_module(name: str):
    try:
        return import_module(name)
    except Exception as exc:
        tb = exc.__traceback__
        tb = tb.tb_next.tb_next  # type: ignore
        while tb and tb.tb_frame.f_code.co_filename.startswith("<frozen importlib"):
            tb = tb.tb_next
        raise exc.with_traceback(tb)


@internal
def macro_call(runtime: Any, function: Any, command: AstCommand):
    with runtime.modules.error_handler(
        f'Macro "{command.identifier}" raised an exception.'
    ):
        return runtime.capture_output(function, *command.arguments)


@internal
def resolve_formatted_location(runtime: Any, nested_path: str) -> str:
    root = runtime.get_nested_location()
    namespace, resolved = resolve_relative_location(
        nested_path,
        root,
        include_root_file=True,
    )
    return f"{namespace}:{resolved}"


def converter(f: Callable[[Any], AstNode]) -> Callable[[Any, AstNode], AstNode]:
    internal(f)

    @internal
    @wraps(f)
    def wrapper(obj: Any, node: AstNode) -> AstNode:
        if isinstance(obj, AstNode):
            return set_location(obj, node)
        return set_location(f(obj), node)

    return wrapper


@dataclass
class JsonObjectConverter:
    """Converter for json objects."""

    json_converter: Callable[[Any, AstNode], AstNode]

    @internal
    def __call__(self, obj: Any, node: AstNode) -> AstNode:
        if isinstance(obj, AstNode):
            return set_location(obj, node)
        if isinstance(node := self.json_converter(obj, node), AstJsonObject):
            return node
        raise ValueError(f"Invalid json object of type {type(obj)!r}.")


@dataclass
class NbtCompoundConverter:
    """Converter for nbt compounds."""

    nbt_converter: Callable[[Any, AstNode], AstNode]

    @internal
    def __call__(self, obj: Any, node: AstNode) -> AstNode:
        if isinstance(obj, AstNode):
            return set_location(obj, node)
        if isinstance(node := self.nbt_converter(obj, node), AstNbtCompound):
            return node
        raise ValueError(f"Invalid nbt compound of type {type(obj)!r}.")


@dataclass
class EntityConverter:
    """Converter for entities."""

    uuid_converter: Callable[[Any, AstNode], AstNode]
    player_name_converter: Callable[[Any, AstNode], AstNode]

    @internal
    def __call__(self, obj: Any, node: AstNode) -> AstNode:
        if isinstance(obj, AstNode):
            return set_location(obj, node)
        if isinstance(obj, str):
            if obj.count("-") == 4:
                return self.uuid_converter(obj, node)
            return self.player_name_converter(obj, node)
        if isinstance(obj, UUID):
            return self.uuid_converter(obj, node)
        raise ValueError(f"Invalid entity value of type {type(obj)!r}.")



__all__ = [
    "CommandEmitter",
]


from contextlib import contextmanager
from typing import Any, Callable, Generator, Iterator, List, Optional, ParamSpec, Tuple

from mecha import AstCommand, AstNode, AstRoot

from .utils import internal

P = ParamSpec("P")


class CommandEmitter:
    """Command emitter."""

    commands: List[AstCommand]
    nesting: List[Tuple[str, Tuple[AstNode, ...]]]

    def __init__(self):
        self.commands = []
        self.nesting = []

    @contextmanager
    def scope(
        self,
        commands: Optional[List[AstCommand]] = None,
    ) -> Iterator[List[AstCommand]]:
        """Create a new scope to gather commands."""
        if commands is None:
            commands = []

        previous_commands = self.commands
        self.commands = commands

        try:
            yield commands
        finally:
            self.commands = previous_commands

    @internal
    def capture_output(
        self,
        f: Callable[P, Any],
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> List[AstCommand]:
        """Invoke a user-defined function and return the list of generated commands."""
        with self.scope() as output:
            result = f(*args, **kwargs)

            if isinstance(result, Generator):
                try:
                    while True:
                        node: Any = next(result)  # type: ignore
                        if isinstance(node, AstRoot):
                            self.commands.extend(node.commands)
                        elif isinstance(node, AstCommand):
                            self.commands.append(node)
                        elif node:
                            msg = f"Invalid command of type {type(node)!r}."
                            raise TypeError(msg)
                except StopIteration as exc:
                    result = exc.value

        if not result:
            result = []
        elif isinstance(result, AstNode):
            result = [result]

        for node in result:
            if isinstance(node, AstRoot):
                output.extend(node.commands)
            elif isinstance(node, AstCommand):
                output.append(node)
            elif node:
                msg = f"Invalid command of type {type(node)!r}."
                raise TypeError(msg)

        return output

    @contextmanager
    def push_nesting(self, identifier: str, *arguments: AstNode):
        self.nesting.append((identifier, arguments))
        try:
            yield
        finally:
            self.nesting.pop()



__all__ = [
    "AstNonFunctionRoot",
    "AstStatement",
    "AstExpression",
    "AstExpressionBinary",
    "AstExpressionUnary",
    "AstChainedComparison",
    "AstValue",
    "AstIdentifier",
    "AstFormatString",
    "AstFormattedLocation",
    "AstTuple",
    "AstList",
    "AstDict",
    "AstDictItem",
    "AstDictUnquotedKey",
    "AstSlice",
    "AstUnpack",
    "AstKeyword",
    "AstAttribute",
    "AstLookup",
    "AstCall",
    "AstTarget",
    "AstTargetIdentifier",
    "AstTargetUnpack",
    "AstTargetAttribute",
    "AstTargetItem",
    "AstAssignment",
    "AstTypeDeclaration",
    "AstDecorator",
    "AstDocstring",
    "AstWithClause",
    "AstWithContext",
    "AstFunctionSignature",
    "AstFunctionSignatureElement",
    "AstFunctionSignatureArgument",
    "AstFunctionSignaturePositionalMarker",
    "AstFunctionSignatureVariadicArgument",
    "AstFunctionSignatureVariadicMarker",
    "AstFunctionSignatureVariadicKeywordArgument",
    "AstDeferredRoot",
    "AstMacro",
    "AstMacroCall",
    "AstMacroLiteral",
    "AstMacroArgument",
    "AstMacroMatch",
    "AstMacroMatchLiteral",
    "AstMacroMatchArgument",
    "AstProcMacro",
    "AstProcMacroMarker",
    "AstProcMacroResult",
    "AstClassName",
    "AstClassBases",
    "AstClassRoot",
    "AstEscapeAnalysisRoot",
    "AstEscapeRoot",
    "AstMemo",
    "AstMemoResult",
    "AstInterpolation",
    "AstFromImport",
    "AstImportedItem",
    "AstImportedMacro",
    "AstPrelude",
]


import re
from dataclasses import dataclass
from typing import Any, List, Optional, Tuple, Union
from uuid import UUID, uuid4

from beet.core.utils import JsonDict, extra_field, required_field
from mecha import (
    AstChildren,
    AstCommand,
    AstCommandSentinel,
    AstJson,
    AstLiteral,
    AstNode,
    AstResourceLocation,
    AstRoot,
)
from tokenstream import TokenStream

from .pattern import IDENTIFIER_PATTERN
from .semantics import Binding


@dataclass(frozen=True, slots=True)
class AstNonFunctionRoot(AstRoot):
    """Non-function root ast node."""


@dataclass(frozen=True, slots=True)
class AstStatement(AstCommandSentinel):
    """Ast statement node."""


@dataclass(frozen=True, slots=True)
class AstExpression(AstNode):
    """Base ast node for expressions."""


@dataclass(frozen=True, slots=True)
class AstExpressionBinary(AstExpression):
    """Ast expression binary node."""

    operator: str = required_field()
    left: AstExpression = required_field()
    right: AstExpression = required_field()


@dataclass(frozen=True, slots=True)
class AstExpressionUnary(AstExpression):
    """Ast expression unary node."""

    operator: str = required_field()
    value: AstExpression = required_field()


@dataclass(frozen=True, slots=True)
class AstChainedComparison(AstExpression):
    """Ast chained comparison node."""

    operators: Tuple[str, ...] = required_field()
    operands: AstChildren[AstExpression] = required_field()


@dataclass(frozen=True, slots=True)
class AstValue(AstExpression):
    """Ast value node."""

    value: Any = required_field()

    @property
    def literal(self) -> str:
        return "..." if self.value is ... else repr(self.value)


@dataclass(frozen=True, slots=True)
class AstIdentifier(AstExpression):
    """Ast identifier node."""

    value: str = required_field()


@dataclass(frozen=True, slots=True)
class AstFormatString(AstExpression):
    """Ast format string node."""

    fmt: str = required_field()
    values: AstChildren[AstExpression] = required_field()


@dataclass(frozen=True, slots=True)
class AstFormattedLocation(AstExpression):
    """Ast formatted location node."""

    fmt: str = required_field()
    values: AstChildren[AstExpression] = required_field()


@dataclass(frozen=True, slots=True)
class AstTuple(AstExpression):
    """Ast tuple node."""

    items: AstChildren[AstExpression] = required_field()


@dataclass(frozen=True, slots=True)
class AstList(AstExpression):
    """Ast list node."""

    items: AstChildren[AstExpression] = required_field()


@dataclass(frozen=True, slots=True)
class AstDictUnquotedKey(AstValue):
    """Ast dict unquoted key node."""


@dataclass(frozen=True, slots=True)
class AstDictItem(AstNode):
    """Ast dict item node."""

    key: AstExpression = required_field()
    value: AstExpression = required_field()


@dataclass(frozen=True, slots=True)
class AstDict(AstExpression):
    """Ast dict node."""

    items: AstChildren[AstDictItem] = required_field()


@dataclass(frozen=True, slots=True)
class AstSlice(AstNode):
    """Ast slice node."""

    start: Optional[AstExpression] = None
    stop: Optional[AstExpression] = None
    step: Optional[AstExpression] = None


@dataclass(frozen=True, slots=True)
class AstUnpack(AstNode):
    """Ast unpack node."""

    type: str = required_field()
    value: AstExpression = required_field()


@dataclass(frozen=True, slots=True)
class AstKeyword(AstNode):
    """Ast keyword node."""

    name: str = required_field()
    value: AstExpression = required_field()


@dataclass(frozen=True, slots=True)
class AstAttribute(AstExpression):
    """Ast attribute node."""

    name: str = required_field()
    value: AstExpression = required_field()


@dataclass(frozen=True, slots=True)
class AstLookup(AstExpression):
    """Ast lookup node."""

    value: AstExpression = required_field()
    arguments: AstChildren[Union[AstExpression, AstSlice]] = required_field()


@dataclass(frozen=True, slots=True)
class AstCall(AstExpression):
    """Ast call node."""

    value: AstExpression = required_field()
    arguments: AstChildren[Union[AstExpression, AstUnpack, AstKeyword]] = (
        required_field()
    )


@dataclass(frozen=True, slots=True)
class AstTarget(AstNode):
    """Base node for targets."""


@dataclass(frozen=True, slots=True)
class AstTargetIdentifier(AstTarget):
    """Ast target identifier node."""

    value: str = required_field()
    rebind: bool = False


@dataclass(frozen=True, slots=True)
class AstTargetUnpack(AstTarget):
    """Ast target unpack node."""

    targets: AstChildren[AstTarget] = required_field()


@dataclass(frozen=True, slots=True)
class AstTargetAttribute(AstTarget):
    """Ast target attribute node."""

    name: str = required_field()
    value: AstExpression = required_field()


@dataclass(frozen=True, slots=True)
class AstTargetItem(AstTarget):
    """Ast target item node."""

    value: AstExpression = required_field()
    arguments: AstChildren[Union[AstExpression, AstSlice]] = required_field()


@dataclass(frozen=True, slots=True)
class AstAssignment(AstNode):
    """Ast assignment node."""

    operator: str = required_field()
    target: AstTarget = required_field()
    value: AstExpression = required_field()
    type_annotation: Optional[AstExpression] = None


@dataclass(frozen=True, slots=True)
class AstTypeDeclaration(AstNode):
    """Ast type declaration node."""

    identifier: AstTargetIdentifier = required_field()
    type_annotation: AstExpression = required_field()


@dataclass(frozen=True, slots=True)
class AstDecorator(AstNode):
    """Ast decorator node."""

    expression: AstExpression = required_field()


@dataclass(frozen=True, slots=True)
class AstDocstring(AstCommandSentinel):
    """Ast docstring node."""


@dataclass(frozen=True, slots=True)
class AstWithClause(AstNode):
    """Ast with clause node."""

    value: AstExpression = required_field()
    target: Optional[AstTarget] = None


@dataclass(frozen=True, slots=True)
class AstWithContext(AstNode):
    """Ast with context node."""

    clauses: AstChildren[AstWithClause] = required_field()


@dataclass(frozen=True, slots=True)
class AstFunctionSignatureElement(AstNode):
    """Base node for function signature element."""


@dataclass(frozen=True, slots=True)
class AstFunctionSignatureArgument(AstFunctionSignatureElement):
    """Ast function signature argument node."""

    name: str = required_field()
    type_annotation: Optional[AstExpression] = None
    default: Optional[AstExpression] = None


@dataclass(frozen=True, slots=True)
class AstFunctionSignaturePositionalMarker(AstFunctionSignatureElement):
    """Ast function signature positional marker node."""


@dataclass(frozen=True, slots=True)
class AstFunctionSignatureVariadicArgument(AstFunctionSignatureElement):
    """Ast function signature variadic argument node."""

    name: str = required_field()
    type_annotation: Optional[AstExpression] = None


@dataclass(frozen=True, slots=True)
class AstFunctionSignatureVariadicMarker(AstFunctionSignatureElement):
    """Ast function signature variadic marker node."""


@dataclass(frozen=True, slots=True)
class AstFunctionSignatureVariadicKeywordArgument(AstFunctionSignatureElement):
    """Ast function signature variadic keyword argument node."""

    name: str = required_field()
    type_annotation: Optional[AstExpression] = None


@dataclass(frozen=True, slots=True)
class AstFunctionSignature(AstNode):
    """Ast function signature node."""

    decorators: AstChildren[AstDecorator] = AstChildren()
    name: str = required_field()
    arguments: AstChildren[AstFunctionSignatureElement] = AstChildren()
    return_type_annotation: Optional[AstExpression] = None


@dataclass(frozen=True, slots=True)
class AstDeferredRoot(AstNode):
    """Ast deferred root node."""

    stream: TokenStream = required_field()


@dataclass(frozen=True, slots=True)
class AstMacro(AstCommand):
    """Ast macro node."""

    def get_command_tree(self) -> JsonDict:
        tree_root: JsonDict = {"type": "root"}
        tree: JsonDict = tree_root

        for node in self.arguments:
            if isinstance(node, AstMacroLiteral):
                child = {"type": "literal"}
                tree["children"] = {node.value: child}
            elif isinstance(node, AstMacroMatchLiteral):
                child = {"type": "literal"}
                tree["children"] = {node.match.value: child}
            elif isinstance(node, AstMacroMatchArgument):
                if node.is_subcommand():
                    tree["redirect"] = (
                        node.match_argument_properties.evaluate().get("redirect", [])
                        if node.match_argument_properties
                        else []
                    )
                    return tree_root
                child = {"type": "argument"}
                child["parser"] = node.match_argument_parser.get_canonical_value()
                if properties := node.match_argument_properties:
                    child["properties"] = properties.evaluate()
                tree["children"] = {node.match_identifier.value: child}
            else:
                break
            tree = child

        tree["executable"] = True

        return tree_root


@dataclass(frozen=True, slots=True)
class AstMacroCall(AstCommand):
    """Ast macro call node."""


@dataclass(frozen=True, slots=True)
class AstMacroLiteral(AstLiteral):
    """Ast macro literal node."""

    parser = "bolt_macro_literal"

    regex = re.compile(r"[^#:\s()]+")


@dataclass(frozen=True, slots=True)
class AstMacroArgument(AstLiteral):
    """Ast macro argument node."""

    parser = "bolt_macro_argument"

    regex = re.compile(IDENTIFIER_PATTERN)


@dataclass(frozen=True, slots=True)
class AstMacroMatch(AstNode):
    """Base node for macro match."""


@dataclass(frozen=True, slots=True)
class AstMacroMatchLiteral(AstMacroMatch):
    """Ast macro match literal node."""

    match: AstMacroLiteral = required_field()


@dataclass(frozen=True, slots=True)
class AstMacroMatchArgument(AstMacroMatch):
    """Ast macro match argument node."""

    match_identifier: AstMacroArgument = required_field()
    match_argument_parser: AstResourceLocation = required_field()
    match_argument_properties: Optional[AstJson] = None

    def is_subcommand(self) -> bool:
        return (
            not self.match_argument_parser.namespace
            and self.match_argument_parser.path == "subcommand"
        )


@dataclass(frozen=True, slots=True)
class AstProcMacro(AstMacro):
    """Ast proc macro node."""


@dataclass(frozen=True, slots=True)
class AstProcMacroMarker(AstNode):
    """Ast proc macro marker node."""


@dataclass(frozen=True, slots=True)
class AstProcMacroResult(AstNode):
    """Ast proc macro result node."""

    commands: AstChildren[AstCommand] = required_field()


@dataclass(frozen=True, slots=True)
class AstClassName(AstNode):
    """Ast class name node."""

    decorators: AstChildren[AstDecorator] = AstChildren()
    value: str = required_field()


@dataclass(frozen=True, slots=True)
class AstClassBases(AstNode):
    """Ast class bases node."""

    inherit: AstChildren[AstExpression] = AstChildren()
    kwargs: AstChildren[AstKeyword] = AstChildren()


@dataclass(frozen=True, slots=True)
class AstClassRoot(AstRoot):
    """Ast class root node."""


@dataclass(frozen=True, slots=True)
class AstEscapeAnalysisRoot(AstRoot):
    """Ast escape analysis root node."""

    refcount_snapshots: List[Tuple[str, Binding, int]] = extra_field(
        default_factory=list
    )


@dataclass(frozen=True, slots=True)
class AstEscapeRoot(AstRoot):
    """Ast escape root node."""

    identifiers: Tuple[str, ...] = ()


@dataclass(frozen=True, slots=True)
class AstMemo(AstCommand):
    """Ast memo node."""

    persistent_id: UUID = extra_field(default_factory=uuid4)


@dataclass(frozen=True, slots=True)
class AstMemoResult(AstCommandSentinel):
    """Ast memo result node."""

    serialized: str = ""

    compile_hints = {"skip_execute_inline_single_command": True}


@dataclass(frozen=True, slots=True)
class AstInterpolation(AstNode):
    """Ast interpolation node."""

    prefix: Optional[str] = None
    unpack: Optional[str] = None
    converter: str = required_field()
    value: AstExpression = required_field()


@dataclass(frozen=True, slots=True)
class AstFromImport(AstCommand):
    """Ast from import node."""

    identifier: str = ""


@dataclass(frozen=True, slots=True)
class AstImportedItem(AstNode):
    """Ast imported item node."""

    name: str = required_field()
    identifier: bool = True


@dataclass(frozen=True, slots=True)
class AstImportedMacro(AstNode):
    """Ast imported macro node."""

    name: str = required_field()
    declaration: AstMacro = required_field()


@dataclass(frozen=True, slots=True)
class AstPrelude(AstFromImport):
    """Ast prelude node."""

    @classmethod
    def placeholder(cls, resource_location: str) -> "AstPrelude":
        return cls(
            arguments=AstChildren(
                [
                    AstResourceLocation.from_value(resource_location),
                    AstImportedItem(name="_bolt_placeholder"),
                ]
            )
        )



__all__ = [
    "Runtime",
    "Evaluator",
    "NonFunctionSerializer",
]


import builtins
from dataclasses import dataclass, field
from functools import partial
from importlib.resources import files
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Union

from beet import Context, TextFileBase, generate_tree
from beet.core.utils import JsonDict, extra_field, required_field
from mecha import (
    AstRoot,
    CommandSpec,
    CommandTree,
    CompilationDatabase,
    CompilationUnitProvider,
    Diagnostic,
    FileTypeCompilationUnitProvider,
    Mecha,
    Visitor,
    rule,
)
from mecha.contrib.nested_location import NestedLocationResolver
from mecha.contrib.relative_location import resolve_relative_location
from pathspec import PathSpec
from tokenstream import set_location

from .ast import AstNonFunctionRoot, AstRoot
from .codegen import Codegen
from .emit import CommandEmitter
from .helpers import get_bolt_helpers
from .loop_info import loop_info
from .memo import MemoHandler, MemoRegistry
from .module import (
    AssetModule,
    CompiledModule,
    Module,
    ModuleCacheBackend,
    ModuleManager,
)
from .parse import get_bolt_parsers
from .utils import internal


class Runtime(CommandEmitter):
    """The bolt runtime."""

    helpers: Dict[str, Any]
    globals: JsonDict
    builtins: Set[str]

    memo: MemoHandler

    modules: ModuleManager
    evaluate: "Evaluator"

    spec: CommandSpec
    module_provider: CompilationUnitProvider

    def __init__(self, ctx: Union[Context, Mecha]):
        super().__init__()
        self.helpers = get_bolt_helpers()
        self.globals = {"_bolt_runtime": self, "ctx": None, "loop_info": loop_info}
        self.builtins = {name for name in dir(builtins) if not name.startswith("_")}

        if isinstance(ctx, Context):
            ctx.require(
                "mecha.contrib.raw",
                "mecha.contrib.relative_location",
                "mecha.contrib.nesting",
                "mecha.contrib.nested_location",
                "mecha.contrib.nested_resources",
                "mecha.contrib.nested_yaml",
                "mecha.contrib.implicit_execute",
                self.finalize,
            )

            ctx.data.extend_namespace.append(Module)
            ctx.assets.extend_namespace.append(AssetModule)

            self.globals["ctx"] = ctx

            self.expose("generate_path", ctx.generate.path)
            self.expose("generate_id", ctx.generate.id)
            self.expose("generate_hash", ctx.generate.hash)
            self.expose("generate_objective", ctx.generate.objective)
            self.expose(
                "generate_tree",
                lambda *args, **kwargs: generate_tree(
                    (
                        root := (
                            kwargs.pop("root")
                            if "root" in kwargs
                            else self.get_nested_location()
                        )
                    ),
                    *args,
                    name=(
                        kwargs.pop("name")
                        if "name" in kwargs
                        else ctx.generate["tree"][root].format("tree_{incr}")
                    ),
                    **kwargs,
                ),
            )

            mc = ctx.inject(Mecha)
            self.memo = MemoHandler(
                mc,
                registry=ctx.inject(MemoRegistry),
                generate=ctx.generate,
            )

        else:
            mc = ctx
            self.memo = MemoHandler(mc, MemoRegistry())

        self.modules = ModuleManager(
            directory=mc.directory,
            database=mc.database,
            codegen=Codegen(),
            parse_callback=mc.parse,
            cache=mc.cache,
            globals=self.globals,
            builtins=self.builtins,
        )

        self.evaluate = Evaluator(modules=self.modules)

        self.spec = mc.spec

        self.module_provider = FileTypeCompilationUnitProvider([Module, AssetModule])
        mc.providers.append(self.module_provider)

        commands_json = files("bolt.resources").joinpath("commands.json").read_text()
        command_tree = CommandTree.parse_raw(commands_json)
        bolt_prototypes = set(CommandSpec(tree=command_tree).prototypes)
        mc.spec.add_commands(command_tree)
        mc.spec.parsers.update(
            get_bolt_parsers(mc.spec.parsers, self.modules, bolt_prototypes)
        )

        mc.steps.insert(0, self.evaluate)

        mc.serialize.extend(NonFunctionSerializer(database=mc.database))
        mc.cache_backend = ModuleCacheBackend(modules=self.modules)

    def expose(self, name: str, function: Callable[..., Any]):
        """Expose a utility function."""
        self.globals[name] = lambda *args, **kwargs: function(*args, **kwargs)  # type: ignore

    @internal
    def import_module(self, resource_location: str) -> CompiledModule:
        """Import module."""
        module = self.modules.get(resource_location)
        if not module:
            raise ImportError(f'Couldn\'t import "{resource_location}".')
        if not module.executing:
            with self.modules.error_handler(
                "Top-level statement raised an exception.", module.resource_location
            ):
                self.modules.eval(module)
        return module

    @internal
    def from_module_import(self, resource_location: str, *args: str) -> Any:
        """Import a specific name from a module."""
        module = self.import_module(resource_location)
        try:
            values = [module.namespace[name] for name in args]
        except KeyError as exc:
            message = f'Couldn\'t import {exc} from "{resource_location}".'
            raise ImportError(message) from None
        return values[0] if len(values) == 1 else values

    def get_nested_location(self) -> str:
        """Return the resource location associated with the current level of nesting."""
        root, relative_path = NestedLocationResolver.concat_nested_path(
            NestedLocationResolver.walk_location_hierarchy(
                self.spec, reversed(self.nesting)
            )
        )

        if not root:
            root = self.modules.current_path

        namespace, resolved = resolve_relative_location(
            relative_path,
            root,
            include_root_file=True,
        )
        return f"{namespace}:{resolved}"

    def finalize(self, ctx: Context):
        """Plugin that runs at the end of the build."""
        try:
            yield
        finally:
            for pack in [ctx.data, *ctx.data.overlays.values()]:
                pack[Module].clear()
            for pack in [ctx.assets, *ctx.assets.overlays.values()]:
                pack[AssetModule].clear()
            self.memo.finalize()


@dataclass
class Evaluator(Visitor):
    """Visitor that evaluates modules."""

    modules: ModuleManager = required_field()

    entrypoint: List[str] = field(default_factory=list)
    entrypoint_spec: PathSpec = extra_field(init=False)

    def __post_init__(self):
        super().__post_init__()
        self.add_entrypoint()

    def add_entrypoint(self, *args: Union[str, Iterable[str]]):
        self.entrypoint.extend(
            entry
            for patterns in args
            for entry in ([patterns] if isinstance(patterns, str) else patterns)
        )
        self.entrypoint_spec = PathSpec.from_lines("gitwildmatch", self.entrypoint)

    @rule(AstRoot)
    @internal
    def root(self, node: AstRoot) -> Optional[AstRoot]:
        compilation_unit, module = self.modules.match_ast(node)

        if (
            isinstance(self.modules.database.current, (Module, AssetModule))
            and not module.executed
            and module.resource_location
            and not self.entrypoint_spec.match_file(module.resource_location)
        ):
            module.execution_hooks.append(
                partial(
                    self.restore_module,
                    self.modules.database.current,
                    node,
                    self.modules.database.step,
                )
            )
            return None

        with self.modules.error_handler(
            "Top-level statement raised an exception.", module.resource_location
        ):
            node = self.modules.eval(module)
            compilation_unit.priority = module.execution_index
            return node

    def restore_module(self, key: TextFileBase[Any], node: AstRoot, step: int):
        compilation_unit = self.modules.database[key]
        compilation_unit.ast = node
        self.modules.database.enqueue(key, step, compilation_unit.priority)


@dataclass
class NonFunctionSerializer(Visitor):
    """Serializer that preserves the original source of non-function files."""

    database: CompilationDatabase = required_field()

    @rule(AstNonFunctionRoot)
    def non_function_root(self, node: AstNonFunctionRoot, result: List[str]):
        if source := self.database[self.database.current].source:
            result.append(source)
        if node.commands:
            command = node.commands[0]
            name = command.identifier.partition(":")[0]
            d = Diagnostic(
                "warn", f'Ignored top-level "{name}" command outside function.'
            )
            return set_location(
                d, command, command.arguments[0] if command.arguments else command
            )



__all__ = [
    "rewrite_traceback",
    "fake_traceback",
    "suggest_typo",
    "load_pickle",
    "dump_pickle",
    "internal",
    "INTERNAL_CODE",
]


import pickle
from bisect import bisect
from difflib import get_close_matches
from types import CodeType, TracebackType
from typing import Any, Iterable, List, Optional, Set, TypeVar

from beet import Container
from beet.core.utils import FileSystemPath
from mecha.utils import string_to_number

T = TypeVar("T")


INTERNAL_CODE: Set[CodeType] = {
    string_to_number.__code__,
    Container[Any, Any].__getitem__.__code__,
}


def internal(f: T) -> T:
    INTERNAL_CODE.add(f.__code__)  # type: ignore
    return f


def rewrite_traceback(exc: Exception) -> Exception:
    tb = exc.__traceback__ and exc.__traceback__.tb_next

    stack: List[TracebackType] = []

    while tb is not None:
        if tb.tb_frame.f_code in INTERNAL_CODE:
            tb = tb.tb_next
            continue

        line_numbers = tb.tb_frame.f_globals.get("_bolt_lineno")

        if line_numbers:
            n1, n2 = line_numbers
            lineno = n2[bisect(n1, tb.tb_lineno) - 1]
            stack.append(fake_traceback(exc, tb, lineno))
        else:
            stack.append(tb)

        tb = tb.tb_next

    tb_next = None

    for tb in reversed(stack):
        tb.tb_next = tb_next
        tb_next = tb

    return exc.with_traceback(tb)


def fake_traceback(exc: Exception, tb: TracebackType, lineno: int) -> TracebackType:  # type: ignore
    name = tb.tb_frame.f_code.co_name
    filename = tb.tb_frame.f_globals["__file__"]

    if name == "<module>":
        name = tb.tb_frame.f_globals.get("__name__")

    code = compile("\n" * (lineno - 1) + "raise _bolt_exc", filename, "exec")

    if name:
        code = code.replace(co_name=name)

    try:
        exec(code, {"_bolt_exc": exc})
    except Exception as exc:
        return exc.__traceback__.tb_next  # type: ignore


def suggest_typo(wrong: str, possibilities: Iterable[str]) -> Optional[str]:
    cutoff = 0.6 if len(wrong) < 3 else 0.7

    if matches := get_close_matches(wrong, possibilities, cutoff=cutoff):
        matches = [f'"{m}"' for m in matches]

        if len(matches) == 1:
            return matches[0]
        else:
            *head, before_last, last = matches
            return ", ".join(head + [f"{before_last} or {last}"])

    return None


def load_pickle(path: FileSystemPath) -> Any:
    with open(path, "rb") as f:
        return pickle.load(f)


def dump_pickle(path: FileSystemPath, value: Any):
    with open(path, "wb") as f:
        return pickle.dump(value, f, protocol=pickle.HIGHEST_PROTOCOL)
